from .api_resource import APIResource

from ..types import DatasetWithItems, Dataset, CreateDatasetOptions


class Datasets(APIResource):
    def load(self, id: str) -> DatasetWithItems:
        resp_data = self._client.request("GET", f"/datasets/{id}")
        return DatasetWithItems(**resp_data["dataset"])

    def list(self) -> list[Dataset]:
        resp_data = self._client.request("GET", "/datasets")
        return [Dataset(**d) for d in resp_data["datasets"]]

    def create(self, create_opts: CreateDatasetOptions) -> Dataset:
        resp_data = self._client.request(
            "POST", "/datasets", json=create_opts.model_dump()
        )
        return Dataset(**resp_data["dataset"])



from collections import defaultdict
from typing import Dict, Any, Optional, List, Union, Sequence
from uuid import UUID
from contextvars import copy_context, Context
import logging
import json

from .. import framework
from ..resources import monitoring
from ..types import GenerationParams, RetrievalParams

try:
    from langchain_core.callbacks.base import BaseCallbackHandler
    from langchain_core.outputs import LLMResult
    from langchain_core.messages import (
        BaseMessage,
        HumanMessage,
        AIMessage,
        SystemMessage,
        ToolMessage,
        FunctionMessage,
        ChatMessage
    )
    from langchain.schema.document import Document
except ImportError:
    raise ModuleNotFoundError(
        "Please install langchain in order to use the Langchain integration."
    )


log = logging.getLogger(__name__)


class LangchainCallbackHandler(BaseCallbackHandler):
    hamming: framework.Hamming
    run_items: Dict[UUID, monitoring.MonitoringItem] = defaultdict(None)
    run_ctx: Dict[UUID, Context] = defaultdict(None)
    run_parent: Dict[UUID, UUID] = defaultdict(None)
    run_llm_input: Dict[UUID, str] = defaultdict(None)
    run_llm_serialized: Dict[UUID, Dict[str, Any]] = defaultdict(None)
    run_retriever_query: Dict[UUID, str] = defaultdict(None)
    run_retriever_serialized: Dict[UUID, Dict[str, Any]] = defaultdict(None)

    def __init__(self, hamming: framework.Hamming):
        self.hamming = hamming
        self.hamming.monitoring.start()
        log.debug("LangchainCallbackHandler initialized!")

    def on_chain_start(
        self, 
        serialized: Dict[str, Any], 
        inputs: Dict[str, Any],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs: Any
    ):
        if parent_run_id is not None:
            self.run_parent[run_id] = parent_run_id
            return
        self.run_items[run_id] = self.hamming.monitoring.start_item(input=inputs)
        self.run_ctx[run_id] = copy_context()

    def on_chain_end(
        self,
        outputs: Union[str, Dict[str, Any]],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any
    ):
        if parent_run_id is not None:
            if run_id in self.run_parent:
                del self.run_parent[run_id]
            return
        
        monitoring_item = self.run_items.pop(run_id)
        if monitoring_item is None:
            log.warning(f"No monitoring item found for run_id: {run_id}")
            return

        if isinstance(outputs, str):
            monitoring_item.set_output({
                "response": outputs
            })
        else:
            monitoring_item.set_output(outputs)
        monitoring_item._end()

    def on_chain_error(
        self,
        error: Exception,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any
    ):
        monitoring_item = self.run_items.pop(run_id)
        if monitoring_item is None:
            log.warning(f"No monitoring item found for run_id: {run_id}")
            return
        monitoring_item.set_output({})
        monitoring_item._end(True, str(error))

    def on_chat_model_start(
        self,
        serialized: Dict[str, Any],
        messages: List[List[BaseMessage]],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ):
        flatten_messages = [message for sublist in messages for message in sublist]
        message_dicts = [self._convert_message(message) for message in flatten_messages]
        self.run_llm_input[run_id] = json.dumps(message_dicts)
        self.run_llm_serialized[run_id] = serialized

    def on_llm_start(
        self,
        serialized: Dict[str, Any],
        prompts: List[str],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Any:
        self.run_llm_input[run_id] = json.dumps(prompts)
        self.run_llm_serialized[run_id] = serialized

    def on_llm_end(
        self,
        response: LLMResult,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        llm_input = self.run_llm_input.pop(run_id)
        llm_serialized = self.run_llm_serialized.pop(run_id)

        metadata=GenerationParams.Metadata()
        if "openai" in llm_serialized.get("id", []):
            metadata.provider = "openai"
            kw = llm_serialized.get("kwargs", {})
            metadata.model = kw.get("model_name", None)
            metadata.temperature = kw.get("temperature", None)
            metadata.max_tokens = kw.get("max_tokens", None)
            metadata.n = kw.get("n", None)

        llm_params = GenerationParams(
            input=llm_input,
            output=response.generations[-1][-1].text,
            metadata=metadata,
        )

        top_parent_run_id = self._find_top_parent_run_id(parent_run_id)
        ctx = self.run_ctx.get(top_parent_run_id, None)
        if ctx is not None:
            ctx.run(self.hamming.tracing.log_generation, llm_params)
        else:
            self.hamming.tracing.log_generation(params=llm_params)

    def on_retriever_start(
        self,
        serialized: Dict[str, Any],
        query: str,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Any:
        self.run_retriever_query[run_id] = query
        self.run_retriever_serialized[run_id] = serialized

    def on_retriever_end(
        self,
        documents: Sequence[Document],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        retriever_query = self.run_retriever_query.pop(run_id)
        retriever_serialized = self.run_retriever_serialized.pop(run_id)
        metadata = RetrievalParams.Metadata(
            engine="langchain" # TODO: detect engine
        )
        retrieval_params = RetrievalParams(
            query=retriever_query,
            results=[document.page_content for document in documents],
            metadata=metadata,
        )
        top_parent_run_id = self._find_top_parent_run_id(parent_run_id)
        ctx = self.run_ctx.get(top_parent_run_id, None)
        if ctx is not None:
            ctx.run(self.hamming.tracing.log_retrieval, retrieval_params)
        else:
            self.hamming.tracing.log_retrieval(params=retrieval_params)

    def _find_top_parent_run_id(self, run_id: UUID):
        parent_run_id = self.run_parent.get(run_id, None)
        if parent_run_id is not None:
            return self._find_top_parent_run_id(parent_run_id)
        return run_id
    

    def _convert_message(self, message: BaseMessage) -> Dict[str, Any]:
        if isinstance(message, HumanMessage):
            message_dict = {"role": "user", "content": message.content}
        elif isinstance(message, AIMessage):
            message_dict = {"role": "assistant", "content": message.content}
        elif isinstance(message, SystemMessage):
            message_dict = {"role": "system", "content": message.content}
        elif isinstance(message, ToolMessage):
            message_dict = {"role": "tool", "content": message.content}
        elif isinstance(message, FunctionMessage):
            message_dict = {"role": "function", "content": message.content}
        elif isinstance(message, ChatMessage):
            message_dict = {"role": message.role, "content": message.content}
        else:
            raise ValueError(f"Got unknown type {message}")
        return message_dict



import asyncio
import inspect
from datetime import datetime
from typing import Optional, Dict, Awaitable
from concurrent.futures import ThreadPoolExecutor

from ..types import (
    DatasetItem,
    Experiment,
    ExperimentItem,
    ExperimentItemContext,
    ExperimentStatus,
    InputType,
    MetadataType,
    OutputType,
    Runner,
    RunOptions,
    RunResult,
    ScoreType,
    TracingMode,
    Score,
    CustomScoringConfig,
)
from .api_resource import APIResource
from .scoring import ScoringHelper
from ..utils import get_url_origin

DEFAULT_SCORE_TYPES: list[ScoreType] = [ScoreType.STRING_DIFF]


class ExperimentItems(APIResource):
    def start(
        self, experiment: Experiment, dataset_item: DatasetItem
    ) -> ExperimentItemContext:
        resp_data = self._client.request(
            "POST",
            f"/experiments/{experiment.id}/items",
            json={"datasetItemId": dataset_item.id, "output": {}, "metrics": {}},
        )
        item = ExperimentItem(**resp_data["item"])
        item_context = ExperimentItemContext(item=item, startTs=datetime.now())
        return item_context

    def end(self, 
        item_context: ExperimentItemContext, 
        output: OutputType,
        scores: Dict[str, Score]
    ):
        item = item_context.item
        start_ts = item_context.startTs
        duration_sec = (datetime.now() - start_ts).total_seconds()
        duration_ms = int(duration_sec * 1000)

        self._client.tracing._flush(item.id)

        self._client.request(
            "PATCH",
            f"/experiments/{item.experimentId}/items/{item.id}",
            json={
                "output": output,
                "scores": {
                    key: value.model_dump() for key, value in scores.items()
                },
                "metrics": {"durationMs": duration_ms},
            },
        )


class Experiments(APIResource):
    _items: ExperimentItems

    @staticmethod
    def generate_name(dataset_name: str) -> str:
        now = datetime.now()
        now.microsecond = 0
        now_str = now.isoformat(sep=" ")
        return f"Experiment for {dataset_name} - {now_str}"

    def __init__(self, client) -> None:
        super().__init__(client)
        self._items = ExperimentItems(client)

    def run(self, opts: RunOptions, run: Runner) -> RunResult:
        self._client.tracing._set_mode(TracingMode.EXPERIMENT)
        
        dataset_id = opts.dataset
        dataset = self._client.datasets.load(dataset_id)

        name = opts.name or Experiments.generate_name(dataset.name)
        scoring = opts.scoring or DEFAULT_SCORE_TYPES
        metadata = opts.metadata or {}

        scoring_helper = ScoringHelper(self._client, scoring)
        scoring_helper.initialize()

        def execute_runner(run: Runner, input: InputType) -> OutputType:
            if inspect.iscoroutinefunction(run):
                return asyncio.run(run(input))
            else:
                return run(input)

        experiment = self._start(
            name, 
            dataset_id, 
            scoring_helper.get_config(), 
            metadata
        )
        url_origin = get_url_origin(self._client.base_url)
        experiment_url = f"{url_origin}/experiments/{experiment.id}"

        def run_item(dataset_item: DatasetItem):
            item_context = self._items.start(experiment, dataset_item)
            output = execute_runner(run, dataset_item.input)
            scores = scoring_helper.score(
                dataset_item.input,
                dataset_item.output,
                output,
            )
            self._items.end(item_context, output, scores)

        try:
            if (opts.parallel):
                if isinstance(opts.parallel, bool):
                    worker_count = None
                elif isinstance(opts.parallel, int):
                    worker_count = opts.parallel
                else:
                    raise ValueError(f"Invalid parallel option: {opts.parallel}")

                with ThreadPoolExecutor(max_workers=worker_count) as executor:
                    futures = [
                        executor.submit(run_item, dataset_item)
                        for dataset_item in dataset.items
                    ]
                    for future in futures:
                        future.result()
            else:
                for dataset_item in dataset.items:
                    run_item(dataset_item)
            self._end(experiment)

            return RunResult(url=experiment_url)
        except Exception as ex:
            self._end(experiment, status=ExperimentStatus.FAILED)
            raise ex
        
    async def arun(self, opts: RunOptions, run: Runner) -> Awaitable[RunResult]:
        self._client.tracing._set_mode(TracingMode.EXPERIMENT)
        
        dataset_id = opts.dataset
        dataset = self._client.datasets.load(dataset_id)

        name = opts.name or Experiments.generate_name(dataset.name)
        scoring = opts.scoring or DEFAULT_SCORE_TYPES
        metadata = opts.metadata or {}

        scoring_helper = ScoringHelper(self._client, scoring)
        scoring_helper.initialize()

        async def execute_runner(run: Runner, input: InputType) -> OutputType:
            if inspect.iscoroutinefunction(run):
                return await run(input)
            else:
                return run(input)

        experiment = self._start(
            name, 
            dataset_id, 
            scoring_helper.get_config(), 
            metadata
        )
        url_origin = get_url_origin(self._client.base_url)
        experiment_url = f"{url_origin}/experiments/{experiment.id}"

        try:
            for dataset_item in dataset.items:
                item_context = self._items.start(experiment, dataset_item)
                output = await execute_runner(run, dataset_item.input)
                scores = scoring_helper.score(
                    dataset_item.input,
                    dataset_item.output,
                    output,
                )
                self._items.end(item_context, output, scores)
            self._end(experiment)
            return RunResult(url=experiment_url)
        except Exception as ex:
            self._end(experiment, status=ExperimentStatus.FAILED)
            raise ex

    def _start(
        self,
        name: str,
        dataset_id: str,
        scoring: list[ScoreType | CustomScoringConfig],
        metadata: MetadataType,
    ) -> Experiment:
        status = ExperimentStatus.RUNNING
        scoring_obj = [
            s.model_dump() if type(s) == CustomScoringConfig else s for s in scoring
        ]
        resp_data = self._client.request(
            "POST",
            "/experiments",
            json={
                "name": name,
                "dataset": dataset_id,
                "status": status,
                "scoring": scoring_obj,
                "metadata": metadata,
            },
        )
        return Experiment(**resp_data["experiment"])

    def _end(
        self,
        experiment: Experiment,
        status: Optional[ExperimentStatus] = ExperimentStatus.FINISHED,
    ):
        self._client.request(
            "PATCH", f"/experiments/{experiment.id}", json={"status": status}
        )



from .api_resource import APIResource
from ..types import (
    ScoreType,
    ScoringFunction,
    CustomScoringConfig,
    InputType,
    OutputType,
    Score,
)

class RegisteredScoringFunction(ScoringFunction):
    registration: CustomScoringConfig

class ScoringHelper(APIResource):
    _initialized: bool = False
    _registered_functions: list[RegisteredScoringFunction]
    _standard_scoring: list[ScoreType]
    _custom_scoring: list[ScoringFunction]

    def __init__(self, client, scoring: list[ScoreType]):
        super().__init__(client)
        self._standard_scoring = [s for s in scoring if type(s) == ScoreType]
        self._custom_scoring = [s for s in scoring if type(s) == ScoringFunction]

    def initialize(self):
        if self._initialized:
            return
        self._register_scoring_functions()
        self._initialized = True

    def get_config(self) -> list[ScoreType | CustomScoringConfig]:
        if not self._initialized:
            raise Exception("ScoringHelper is not initialized")
        return self._standard_scoring + [f.registration for f in self._registered_functions]
    
    def score(self, input: InputType, expected: OutputType, actual: OutputType) -> dict[str, Score]:
        if not self._initialized:
            raise Exception("ScoringHelper is not initialized")
        scores = {}
        for f in self._registered_functions:
            score = f.scorer.score_fn({ 
                "input": input, 
                "expected": expected, 
                "output": actual 
            })
            scores[f.registration.key_name] = score
        return scores
    
    def _register_scoring_functions(self):
        scoring_obj = [{
            "name": f.name,
            "version": f.version,
            "score_config": f.score_config.model_dump(),
            "execution_config": {},
        } for f in self._custom_scoring]

        data = self._client.request(
            "POST",
            "/scoring/register-functions",
            json={
                "scoring": scoring_obj
            }
        )
        registrations = data["scoring"]
        self._registered_functions = [
            RegisteredScoringFunction(
                **f.model_dump(),
                registration=registrations[idx]
            ) 
        for idx, f in enumerate(self._custom_scoring)]



import asyncio
from typing import TypeVar, ParamSpec, Callable, Any, Dict, Optional, Mapping
import inspect
import logging

from hamming import get_client


logger = logging.getLogger(__name__)

R = TypeVar("R", covariant=True)
P = ParamSpec("P")

def is_async(func: Callable[P, R]) -> bool:
    return asyncio.iscoroutinefunction(func)

def traced(
    *, 
    name: Optional[str] = None, 
    metadata: Optional[Mapping[str, Any]] = None
):
    def decorator(func: Callable[P, R]) -> Callable[P, R]:
        def inner(*args: Any, **kwargs: Any):
            client = get_client()
            signature = inspect.signature(func)
            inputs = _get_inputs(signature, *args, **kwargs)
            name_ = name or func.__name__
            with client.monitoring.start_item(
                inputs, name=name_, metadata=metadata
            ) as monitoring_item:
                result = func(*args, **kwargs)
                outputs = result if isinstance(result, dict) else {"result": result}
                monitoring_item.set_output(outputs)
            return result
        
        async def async_inner(*args: Any, **kwargs: Any):
            client = get_client()
            signature = inspect.signature(func)
            inputs = _get_inputs(signature, *args, **kwargs)
            name_ = name or func.__name__
            with client.monitoring.start_item(
                inputs, name=name_, metadata=metadata
            ) as monitoring_item:
                result = await func(*args, **kwargs)
                outputs = result if isinstance(result, dict) else {"result": result}
                monitoring_item.set_output(outputs)
            return result
        
        if is_async(func):
            return async_inner
        else:
            return inner
    return decorator

def _get_inputs(
    sig: inspect.Signature, 
    *args: Any, 
    **kwargs: Any
) -> Dict[str, Any]:
    try:
        bound_args = sig.bind_partial(*args, **kwargs)
        bound_args.apply_defaults()
        arguments = dict(bound_args.arguments)
        arguments.pop("self", None)
        arguments.pop("cls", None)
        for name, param in sig.parameters.items():
            if param.kind == inspect.Parameter.VAR_KEYWORD:
                if name in arguments:
                    arguments.update(arguments[name])
                    arguments.pop(name)
        return arguments
    except BaseException as e:
        logger.debug(f"Error getting inputs for signature {sig}: {e}")
        return {}



import logging
from typing import List

from . import APIResource
from ..types import (
    ExperimentTrace,
    MonitoringTrace,
    TraceEventType,
    GenerationParams,
    RetrievalParams,
    Document,
    LogMessage,
    LogMessageType,
    TracingMode,
)

log = logging.getLogger(__name__)


class Tracing(APIResource):
    _collected_events: List[TraceEventType] = []
    _current_local_trace_id: int = 0

    _mode: TracingMode = TracingMode.OFF

    def _set_mode(self, mode: TracingMode):
        self._mode = mode

    def _next_trace_id(self) -> int:
        self._current_local_trace_id += 1
        return self._current_local_trace_id

    def _flush(self, experiment_item_id: str):
        if self._mode != TracingMode.EXPERIMENT:
            log.warning("Tracing mode must be set to <experiment>!")
            return
        events = self._collected_events
        self._collected_events = []

        root_trace = ExperimentTrace(
            id=self._next_trace_id(),
            experimentItemId=experiment_item_id,
            event={"kind": "root"},
        )

        traces: List[ExperimentTrace] = [root_trace]
        for event in events:
            traces.append(
                ExperimentTrace(
                    id=self._next_trace_id(),
                    experimentItemId=experiment_item_id,
                    parentId=root_trace.id,
                    event=event,
                )
            )

        # TODO: Remove exclude_none when the API supports it
        # It will be a breaking change.
        trace_objects = [t.model_dump(exclude_none=True) for t in traces]
        self._client.request("POST", "/traces", json={"traces": trace_objects})

    @staticmethod
    def _generation_event(params: GenerationParams) -> TraceEventType:
        event = params.model_dump(exclude_none=True)
        event["kind"] = "llm"
        return event

    @staticmethod
    def _retrieval_event(params: RetrievalParams) -> TraceEventType:
        def normalize_document(doc: Document | str) -> Document:
            if isinstance(doc, str):
                return Document(pageContent=doc, metadata={})
            return doc

        params.results = [normalize_document(r) for r in params.results]
        event = params.model_dump(exclude_none=True)
        event["kind"] = "vector"
        return event

    def _log_live_trace(self, trace: MonitoringTrace):
        if self._mode != TracingMode.MONITORING:
            log.warning("Tracing mode must be set to <monitoring>!")
            return
        self._client._logger.log(
            LogMessage(type=LogMessageType.Monitoring, payload=trace)
        )

    def log(self, trace: TraceEventType):
        if self._mode == TracingMode.MONITORING:
            context = self._client.monitoring._get_tracing_context()
            self._log_live_trace(
                MonitoringTrace(
                    session_id=context.session_id,
                    seq_id=context.seq_id,
                    parent_seq_id=context.parent_seq_id,
                    root_seq_id=context.root_seq_id,
                    event=trace,
                )
            )
        elif self._mode == TracingMode.EXPERIMENT:
            self._collected_events.append(trace)
        else:
            log.warning("Attempt to send a log trace, but tracing mode is off!")

    def log_generation(self, params: GenerationParams):
        self.log(Tracing._generation_event(params))

    def log_retrieval(self, params: RetrievalParams):
        self.log(Tracing._retrieval_event(params))



from typing import Optional

from ..types import Prompt, FullPromptContent
from .api_resource import APIResource

class Prompts(APIResource):
    def list(self) -> list[Prompt]:
        resp_data = self._client.request("GET", "/prompts")
        prompts = [Prompt(**prompt_data) for prompt_data in resp_data["prompts"]]
        return prompts

    def get(self, slug: str, label: Optional[str] = None, version: Optional[str] = None) -> Prompt:
        url = f"/prompts/{slug}"
        if label:
            url += f"?label={label}"
        if version:
            url += f"&version={version}"
        resp_data = self._client.request("GET", url)
        prompt = FullPromptContent(**resp_data["prompt"])
        return prompt



from datetime import datetime
from enum import Enum
from pydantic import BaseModel
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    Mapping,
    Optional,
    TypeAlias,
    Union,
)


class ClientOptions(BaseModel):
    api_key: str
    base_url: Optional[str] = None


class HttpClientOptions(BaseModel):
    api_key: str
    base_url: str


class ScoreType(str, Enum):
    ACCURACY_AI = "accuracy_ai"
    FACTS_COMPARE = "facts_compare"
    CONTEXT_RECALL = "context_recall"
    CONTEXT_PRECISION = "context_precision"
    HALLUCINATION = "hallucination"
    STRING_DIFF = "string_diff"
    REFUSAL = "refusal"
    SQL_AST = "sql_ast"


InputType: TypeAlias = Dict
OutputType: TypeAlias = Dict
MetadataType: TypeAlias = Dict


Runner: TypeAlias = Union[
    Callable[[InputType], OutputType], Callable[[InputType], Awaitable[OutputType]]
]


class RunResult(BaseModel):
    url: str


class DatasetItemValue(BaseModel):
    input: InputType
    output: OutputType
    metadata: MetadataType


class DatasetItem(DatasetItemValue):
    id: str


class Dataset(BaseModel):
    id: str
    name: str
    description: Optional[str]


class DatasetWithItems(Dataset):
    items: list[DatasetItem]


class CreateDatasetOptions(BaseModel):
    name: str
    description: Optional[str]
    items: list[DatasetItemValue]


class ExperimentStatus(str, Enum):
    CREATED = "CREATED"
    RUNNING = "RUNNING"
    SCORING = "SCORING"
    SCORING_FAILED = "SCORING_FAILED"
    FINISHED = "FINISHED"
    FAILED = "FAILED"


class ExperimentItemStatus(str, Enum):
    CREATED = "CREATED"
    SCORING = "SCORING"
    SCORED = "SCORED"
    SCORING_FAILED = "SCORING_FAILED"
    FAILED = "FAILED"


class Experiment(BaseModel):
    id: str
    name: str
    description: Optional[str]
    datasetId: str
    datasetVersionId: Optional[int]
    status: ExperimentStatus


class ExperimentItemMetrics(BaseModel):
    durationMs: Optional[int] = 0


class ExperimentItem(BaseModel):
    id: str
    experimentId: str
    datasetItemId: str
    output: OutputType
    metrics: ExperimentItemMetrics
    status: ExperimentItemStatus


class ExperimentItemContext(BaseModel):
    item: ExperimentItem
    startTs: datetime


TraceEventType: TypeAlias = Mapping[str, Any]


class LLMProvider(str, Enum):
    OPENAI = "openai"
    ANTHROPIC = "anthropic"
    AZURE_OPENAI = "azure_openai"


class GenerationParams(BaseModel):
    class Usage(BaseModel):
        completion_tokens: int
        prompt_tokens: int
        total_tokens: int

    class Metadata(BaseModel):
        provider: Optional[Union[LLMProvider, str]] = None
        model: Optional[str] = None
        stream: Optional[bool] = None
        max_tokens: Optional[int] = None
        n: Optional[int] = None
        seed: Optional[int] = None
        temperature: Optional[float] = None
        usage: Optional["GenerationParams.Usage"] = None
        duration_ms: Optional[int] = None
        error: bool = False
        error_message: Optional[str] = None

    input: Optional[str] = None
    output: Optional[str] = None
    metadata: Optional[Metadata] = None


class Document(BaseModel):
    pageContent: str
    metadata: Mapping[str, Any]


class RetrievalParams(BaseModel):
    class Metadata(BaseModel):
        engine: Optional[str] = None

    query: Optional[str] = None
    results: Union[list[Document], list[str]]
    metadata: Optional[Metadata] = None


class ExperimentTrace(BaseModel):
    id: int
    experimentItemId: str
    parentId: Optional[int] = None
    event: TraceEventType


class MonitoringTraceContext(BaseModel):
    session_id: str
    seq_id: int
    parent_seq_id: Optional[int] = None
    root_seq_id: Optional[int] = None


class MonitoringTrace(MonitoringTraceContext):
    event: TraceEventType


class MonitoringItemStatus(str, Enum):
    STARTED = "STARTED"
    COMPLETED = "COMPLETED"
    FAILED = "FAILED"


class LogMessageType(int, Enum):
    Monitoring = 1


class LogMessage(BaseModel):
    type: LogMessageType
    payload: MonitoringTrace

class TracingMode(Enum):
    OFF = "off"
    MONITORING = "monitoring"
    EXPERIMENT = "experiment"

class Score(BaseModel):
    value: float
    reason: Optional[str]

class FunctionType(str, Enum):
    NUMERIC = "numeric"
    CLASSIFICATION = "classification"

class FunctionAggregateType(str, Enum):
    MEAN = "mean"
    MEDIAN = "median"

class LabelColor(str, Enum):
    GRAY = "gray"
    LIGHT_GREEN = "light-green"
    LIGHT_BLUE = "light-blue"
    AMBER = "amber"
    PURPLE = "purple"
    PINK = "pink"
    GREEN = "green"
    PASTEL_GREEN = "pastel-green"
    YELLOW = "yellow"
    BLUE = "blue"
    RED = "red"

class NumericScoreConfig(BaseModel):
    type: FunctionType = FunctionType.NUMERIC
    aggregate: FunctionAggregateType

class ClassificationScoreConfig(BaseModel):
    type: FunctionType = FunctionType.CLASSIFICATION
    labels: Dict[int, str]
    colors: Optional[Dict[int, LabelColor]]

ScoreConfig = Union[NumericScoreConfig, ClassificationScoreConfig]


class ScorerExecutionType(str, Enum):
    LOCAL = "local"
    REMOTE = "remote"

class ScoreArgs(BaseModel):
    input: InputType
    output: OutputType
    expected: OutputType

class LocalScorer(BaseModel):
    type: ScorerExecutionType = ScorerExecutionType.LOCAL
    score_fn: Callable[[ScoreArgs], Score]

Scorer = LocalScorer

class ScoringFunction(BaseModel):
    name: str
    version: int
    score_config: ScoreConfig
    scorer: Scorer

class CustomScoringConfig(BaseModel):
    id: str
    key_name: str


class RunOptions(BaseModel):
    dataset: str
    name: Optional[str]
    scoring: Optional[list[ScoreType | ScoringFunction]]
    metadata: Optional[MetadataType]
    parallel: Optional[bool | int] = False

class Prompt(BaseModel):
    slug: str

class ToolChoice(BaseModel):
    choice: str
    functionName: Optional[str] = None

class PromptSettings(BaseModel):
    temperature: Optional[float] = None
    maxTokens: Optional[int] = None
    topP: Optional[float] = None
    frequencyPenalty: Optional[float] = None
    presencePenalty: Optional[float] = None
    toolChoice: Optional[ToolChoice] = None

class ChatMessage(BaseModel):
    role: str
    content: str

class PromptContent(BaseModel):
    languageModel: str
    promptSettings: PromptSettings
    chatMessages: list[ChatMessage] = []
    tools: Optional[Dict[str, str]] = None

class FullPromptContent(Prompt):
    content: PromptContent



import time
from typing import Optional
from json import dumps

from .framework import Hamming
from .types import GenerationParams, LLMProvider


class OpenAILogger:
    from openai.types.chat import ChatCompletion, ChatCompletionChunk

    def __init__(self, hamming_client: Hamming):
        self._client = hamming_client

    def log_chat_completion_error(
        self,
        req_kwargs: dict,
        error_message: str,
    ):
        self._log_chat_completion(
            req_kwargs=req_kwargs,
            error=True,
            error_message=error_message,
        )

    def log_chat_completion_success(
        self,
        req_kwargs: dict,
        duration_ms: int,
        resp: ChatCompletion,
    ):
        self._log_chat_completion(
            req_kwargs=req_kwargs,
            duration_ms=duration_ms,
            resp=resp,
        )
    
    def log_chat_completion_success_streaming(
        self,
        req_kwargs: dict,
        duration_ms: int,
        resp_chunks: list[ChatCompletionChunk],
    ):
        self._log_chat_completion(
            req_kwargs=req_kwargs,
            duration_ms=duration_ms,
            resp_chunks=resp_chunks,
        )

    def _log_chat_completion(
        self,
        req_kwargs: dict,
        duration_ms: int = 0,
        resp: Optional[ChatCompletion] = None,
        resp_chunks: Optional[list[ChatCompletionChunk]] = None,
        error: bool = False,
        error_message: Optional[str] = None,
    ):
        model = resp.model if resp else req_kwargs.get("model")
        req_msgs = req_kwargs.get("messages")
        if resp:
            resp_msgs = [c.model_dump() for c in resp.choices]
        elif resp_chunks:
            resp_msgs = [[c.model_dump() for c in chunk.choices] for chunk in resp_chunks]     
        else:
            resp_msgs = []
        self._client.tracing.log_generation(
            GenerationParams(
                input=dumps(req_msgs),
                output=dumps(resp_msgs),
                metadata=GenerationParams.Metadata(
                    provider=LLMProvider.OPENAI,
                    model=model,
                    stream=req_kwargs.get("stream"),
                    max_tokens=req_kwargs.get("max_tokens"),
                    n=req_kwargs.get("n"),
                    seed=req_kwargs.get("seed"),
                    temperature=req_kwargs.get("temperature"),
                    usage=(
                        GenerationParams.Usage(**resp.usage.model_dump())
                        if resp
                        else None
                    ),
                    duration_ms=duration_ms,
                    error=error,
                    error_message=error_message,
                ),
            )
        )


class WrappedSyncCompletions:
    def __init__(self, completions, logger: OpenAILogger):
        self.__original = completions
        self._logger = logger

    def __getattr__(self, name):
        return getattr(self.__original, name)

    def create(self, *args, **kwargs):
        start_ts = time.time()
        is_stream = kwargs.get("stream", False)
        try:
            resp = self.__original.create(*args, **kwargs)
            if is_stream:
                def gen():
                    all_chunks = []
                    try:
                        for chunk in resp:
                            all_chunks.append(chunk)
                            yield chunk
                    except Exception as e:
                        self._logger.log_chat_completion_error(
                            req_kwargs=kwargs,
                            error_message=str(e),
                        )
                        raise e
                    self._logger.log_chat_completion_success_streaming(
                        req_kwargs=kwargs,
                        duration_ms=int((time.time() - start_ts) * 1000),
                        resp_chunks=all_chunks,
                    )
                return gen()
            else:                
                self._logger.log_chat_completion_success(
                    req_kwargs=kwargs,
                    duration_ms=int((time.time() - start_ts) * 1000),
                    resp=resp,
                )
                return resp
        except Exception as e:
            self._logger.log_chat_completion_error(
                req_kwargs=kwargs,
                error_message=str(e),
            )
            raise e


class WrappedAsyncCompletions:
    def __init__(self, completions, logger: OpenAILogger):
        self.__original = completions
        self._logger = logger

    def __getattr__(self, name):
        return getattr(self.__original, name)

    async def create(self, *args, **kwargs):
        start_ts = time.time()
        is_stream = kwargs.get("stream", False)
        try:
            resp = await self.__original.create(*args, **kwargs)
            if is_stream:
                async def gen():
                    all_chunks = []
                    try:
                        async for chunk in resp:
                            all_chunks.append(chunk)
                            yield chunk
                    except Exception as e:
                        self._logger.log_chat_completion_error(
                            req_kwargs=kwargs,
                            error_message=str(e),
                        )
                        raise e
                    self._logger.log_chat_completion_success_streaming(
                        req_kwargs=kwargs,
                        duration_ms=int((time.time() - start_ts) * 1000),
                        resp_chunks=all_chunks,
                    )
                return gen()
            else:                
                self._logger.log_chat_completion_success(
                    req_kwargs=kwargs,
                    duration_ms=int((time.time() - start_ts) * 1000),
                    resp=resp,
                )
                return resp
        except Exception as e:
            self._logger.log_chat_completion_error(
                req_kwargs=kwargs,
                error_message=str(e),
            )
            raise e


def wrap_openai(openai_client, hamming_client):
    from openai import OpenAI, AsyncOpenAI

    logger = OpenAILogger(hamming_client)

    def _instrument_sync_client(openai_client: OpenAI):
        openai_client.chat.completions = WrappedSyncCompletions(
            openai_client.chat.completions, logger
        )

    def _instrument_async_client(openai_client: AsyncOpenAI):
        openai_client.chat.completions = WrappedAsyncCompletions(
            openai_client.chat.completions, logger
        )

    if isinstance(openai_client, OpenAI):
        _instrument_sync_client(openai_client)
    elif isinstance(openai_client, AsyncOpenAI):
        _instrument_async_client(openai_client)
    else:
        print("Unknown client type.")
    return openai_client



import requests
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry
from typing import Any

from .types import HttpClientOptions


MAX_RETRY_COUNT = 5


class HttpClient:
    api_key: str
    base_url: str
    session: requests.Session

    @staticmethod
    def sanitize_base_url(url: str) -> str:
        return url.strip().rstrip("/")

    @staticmethod
    def create_session() -> requests.Session:
        session = requests.Session()
        retry_strategy = Retry(
            total=MAX_RETRY_COUNT,
            allowed_methods=["GET", "POST", "DELETE", "PATCH", "HEAD", "OPTIONS"],
            backoff_factor=0.5,
            backoff_jitter=1,
            status_forcelist=[429, 500, 502, 503, 504],
        )
        adapter = HTTPAdapter(max_retries=retry_strategy)
        session.mount("http://", adapter)
        session.mount("https://", adapter)
        return session

    def __init__(self, config: HttpClientOptions) -> None:
        self.api_key = config.api_key
        self.base_url = HttpClient.sanitize_base_url(config.base_url)
        self.session = HttpClient.create_session()

    def request(self, method: str, url: str, **kwargs) -> Any:
        resp = self.session.request(
            method=method,
            url=self.base_url + url,
            headers={
                "Content-Type": "application/json",
                "Authorization": f"Bearer {self.api_key}",
            },
            **kwargs,
        )
        resp.raise_for_status()
        if resp.content:
            return resp.json()
        else:
            return None



from .types import ClientOptions, HttpClientOptions
from .http_client import HttpClient
from . import resources


DEFAULT_BASE_URL = "https://app.hamming.ai/api/rest"


class Hamming(HttpClient):
    experiments: resources.Experiments
    datasets: resources.Datasets
    tracing: resources.Tracing
    monitoring: resources.Monitoring
    prompts: resources.Prompts
    
    _logger: resources.AsyncLogger

    def __init__(self, config: ClientOptions) -> None:
        super().__init__(
            HttpClientOptions(
                api_key=config.api_key, base_url=config.base_url or DEFAULT_BASE_URL
            )
        )
        self.experiments = resources.Experiments(self)
        self.datasets = resources.Datasets(self)
        self.tracing = resources.Tracing(self)
        self.monitoring = resources.Monitoring(self)
        self.prompts = resources.Prompts(self)

        self._logger = resources.AsyncLogger(self)
        self._logger.start()
        _set_client(self)

def get_client() -> Hamming:
    global _CLIENT
    if _CLIENT is None:
        raise ValueError("Hamming client not initialized")
    return _CLIENT

def _set_client(client: Hamming) -> None:
    global _CLIENT
    _CLIENT = client



from dataclasses import dataclass
import contextvars
from typing import Optional
import uuid
import threading
import time
import logging

from .api_resource import APIResource
from ..types import (
    MonitoringTrace, 
    MonitoringTraceContext, 
    MonitoringItemStatus, 
    TracingMode
)

logger = logging.getLogger(__name__)


@dataclass
class Session:
    id: str = str(uuid.uuid4())
    seq_id: int = 0


class Monitoring(APIResource):
    _lock = threading.RLock()
    _session: Optional[Session] = None
    _current_tree = contextvars.ContextVar("hamming_monitoring_tree", default=None)

    def start(self):
        with self._lock:
            if not self._session:
                self._session = Session()
            self._client.tracing._set_mode(TracingMode.MONITORING)

    def stop(self):
        with self._lock:
            self._session = None
            self._client.tracing._set_mode(TracingMode.OFF)

    def start_item(
        self, 
        input: Optional[dict] = {}, 
        name: Optional[str] = None, 
        metadata: Optional[dict] = {}
    ):
        (session_id, seq_id) = self._next_seq_id()
        current_tree = self._current_tree.get()
        if not current_tree:
            current_tree = MonitoringTree()
        parent_item = current_tree.get_current_item()
        parent_seq_id = parent_item._seq_id if parent_item else None
        root_item = current_tree.get_root_item()
        root_seq_id = root_item._seq_id if root_item else None
        item = MonitoringItem(
            self, 
            session_id, 
            seq_id, 
            parent_seq_id, 
            root_seq_id
        )
        item._start(input, name, metadata)
        current_tree.add_child(item)
        self._current_tree.set(current_tree)
        return item

    def _end_item(self, trace: MonitoringTrace):
        current_tree: MonitoringTree = self._current_tree.get()
        if not current_tree:
            logger.warn("Missing monitoring tree")
        else:
            try:
                current_tree.pop_child()
            except IndexError:
                logger.warn("Missing monitoring item")
        self._client.tracing._log_live_trace(trace)

    def _get_tracing_context(self) -> MonitoringTraceContext:
        with self._lock:
            if not self._session:
                raise Exception("Monitoring not started")
            current_tree: MonitoringTree = self._current_tree.get()
            current_item = None
            root_item = None
            if current_tree:
                current_item = current_tree.get_current_item()
                root_item = current_tree.get_root_item()
            (session_id, seq_id) = self._next_seq_id()
            return MonitoringTraceContext(
                session_id=session_id,
                seq_id=seq_id,
                parent_seq_id=current_item._seq_id if current_item else None,
                root_seq_id=root_item._seq_id if root_item else None,
            )

    def _next_seq_id(self) -> tuple[str, int]:
        with self._lock:
            if not self._session:
                raise Exception("Monitoring not started")
            self._session.seq_id += 1
            return (self._session.id, self._session.seq_id)


class MonitoringItem:
    _session_id: str
    _seq_id: int
    _parent_seq_id: Optional[int]
    _root_seq_id: Optional[int]
    _input: Optional[dict] = None
    _output: Optional[dict] = None
    _name: Optional[str] = None
    _metadata: Optional[dict] = None
    _metrics: dict
    _status: MonitoringItemStatus
    _error_message: Optional[str] = None
    _start_ts: Optional[float]

    def __init__(
        self, 
        monitoring: Monitoring, 
        session_id: str, 
        seq_id: int, 
        parent_seq_id: Optional[int] = None,
        root_seq_id: Optional[int] = None
    ):
        self._monitoring = monitoring
        self._session_id = session_id
        self._seq_id = seq_id
        self._parent_seq_id = parent_seq_id
        self._root_seq_id = root_seq_id
        self._metrics = {}

    def set_input(self, input: dict):
        self._input = input

    def set_output(self, output: dict):
        self._output = output

    def set_metadata(self, metadata: dict):
        self._metadata = metadata

    def _start(
        self, 
        input: Optional[dict] = {}, 
        name: Optional[str] = None, 
        metadata: Optional[dict] = {}
    ):
        self._input = input
        self._name = name
        self._metadata = metadata
        self._start_ts = time.time()
        self._status = MonitoringItemStatus.STARTED

    def _end(self, error: bool = False, error_message: Optional[str] = None):
        if self._has_ended():
            return
        self._metrics["duration_ms"] = int((time.time() - self._start_ts) * 1000)
        self._status = (
            MonitoringItemStatus.FAILED if error else MonitoringItemStatus.COMPLETED
        )
        self._error_message = error_message
        self._monitoring._end_item(self._to_trace())

    def _has_ended(self) -> bool:
        return self._status in [
            MonitoringItemStatus.COMPLETED,
            MonitoringItemStatus.FAILED,
        ]

    def _to_trace(self) -> MonitoringTrace:
        event_kind = "root" if self._parent_seq_id is None else "span"
        return MonitoringTrace(
            session_id=self._session_id,
            seq_id=self._seq_id,
            parent_seq_id=self._parent_seq_id,
            root_seq_id=self._root_seq_id,
            event={
                "kind": event_kind,
                "input": self._input,
                "output": self._output,
                "name": self._name,
                "metadata": self._metadata,
                "metrics": self._metrics,
                "status": self._status,
                "error_message": self._error_message,
            },
        )

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type:
            self._end(error=True, error_message=exc_value)
        else:
            self._end()


class MonitoringTree:
    path: list[MonitoringItem]

    def __init__(self):
        self.path = []

    def add_child(self, item: MonitoringItem):
        self.path.append(item)

    def pop_child(self):
        return self.path.pop()
    
    def get_current_item(self) -> Optional[MonitoringItem]:
        return self.path[-1] if self.path and len(self.path) > 0 else None

    def get_root_item(self) -> Optional[MonitoringItem]:
        return self.path[0] if self.path and len(self.path) > 0 else None



from __future__ import annotations

import atexit
import logging
import queue
import threading
from typing import TYPE_CHECKING

from ..types import LogMessage
from .api_resource import APIResource

if TYPE_CHECKING:
    from .. import framework


LOG_BATCH_SIZE = 512


log = logging.getLogger(__name__)


class AsyncLogger(APIResource):
    _queue: queue.Queue[LogMessage]
    _thread: threading.Thread
    _stop_event: threading.Event

    def __init__(self, _client: framework.Hamming):
        super().__init__(_client)
        self._queue = queue.Queue()
        self._thread = threading.Thread(target=self._run_queue, daemon=True)
        self._stop_event = threading.Event()
        atexit.register(self.stop)

    def log(self, message: LogMessage):
        self._queue.put(message)

    def start(self):
        log.info("Starting logger thread..")
        self._stop_event.clear()
        self._thread.start()

    def stop(self):
        log.info("Waiting for logger thread to exit..")
        self._stop_event.set()
        self._queue.put(None)
        self._thread.join()
        log.info("Logger thread exited!")

    def _drain_queue(self):
        drained_msgs = []
        while self._queue.qsize() > 0 and len(drained_msgs) < LOG_BATCH_SIZE:
            try:
                msg = self._queue.get_nowait()
                if msg is not None:
                    drained_msgs.append(msg)
            except queue.Empty:
                break
        return drained_msgs

    def _process_queue(self) -> bool:
        log.debug("Processing loop started..")

        msg = self._queue.get()
        if msg is None:
            log.info("Received stop signal from queue!")
            self._queue.task_done()
            return True
        msgs_to_process = [msg]
        msgs_to_process.extend(self._drain_queue())

        log.debug(f"Processing {len(msgs_to_process)} message(s) from the queue")

        self._publish(msgs_to_process)
        for msg in msgs_to_process:
            self._queue.task_done()
        log.debug("Processing loop done!")
        return False

    def _run_queue(self):
        while not self._stop_event.is_set():
            done = self._process_queue()
            if done:
                # Skip last loop if we're shutting down
                return
        # Last processing loop to drain the queue
        self._process_queue()

    def _publish(self, msgs: list[LogMessage]):
        log.debug(f"Publishing {len(msgs)} messages..")
        try:
            logs = [m.model_dump(exclude_none=True) for m in msgs]
            self._client.request("POST", "/logs", json={"logs": logs})
            log.debug(f"Published {len(msgs)} messages!")
        except Exception as e:
            log.error(f"Failed to publish messages: {e}")



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .. import framework


class APIResource:
    _client: framework.Hamming

    def __init__(self, client: framework.Hamming):
        self._client = client
